/* Copyright (c) 2013 APDM
*
* mobility_clinic_main.js
*/

// Set the paths for server host and port
/* global utils:false, longitudinalPlot:false */
var serverInfo = utils.parseURLParams(window.location.href);

// Debug flags
var PLOT_DEBUG = false;

// Values for the "Back" and "Next" buttons
var back = null;
var next = null;

// Global name for unassigned subjects
var UNASSIGNED_SUBJECT = "Unnamed";

// Start a poll on the server to get status information
var polling = false;
var serverStatus = null;
// Start the polling
setInterval(pollServer, 1000);

// Enum for error types
var errorType = {
    DOCK_ERROR: 0,
    MONITOR_ERROR: 1,
    CONFIGURATION_ERROR: 2,
    FILE_ERROR: 3,
    SYNC_ERROR: 4
};

// Is the error page being displayed?
var errorPageOpen = false;

// Is the record mode streaming
var isStreaming = false;

// Constants to use when posting state back to server
// IMPORTANT! These must match the names in
// com.apdm.mobilitylab.views.ClinicView.ClinicUIState
var WAITING_FOR_INIT = "WAITING_FOR_INIT";
var WAITING_FOR_HW = "WAITING_FOR_HW";
var WAITING_FOR_ATTACH = "WAITING_FOR_ATTACH";
var WAITING_FOR_DOCK = "WAITING_FOR_DOCK";
var RUNNING_TEST = "RUNNING_TEST";
var WAITING_FOR_ANALYSIS = "WAITING_FOR_ANALYSIS";
var READING_RESULTS = "READING_RESULTS";
var RESTARTING_TEST = "RESTARTING_TEST";
var DOCK_ERROR_PAGE = "DOCK_ERROR_PAGE";
var CONFIGURATION_ERROR_PAGE = "CONFIGURATION_ERROR_PAGE";
var FILE_ERROR_PAGE = "FILE_ERROR_PAGE";
var SYNC_ERROR_PAGE = "SYNC_ERROR_PAGE";

// Supported trial type metrics
var gaitTrackMetrics = null;
var iswayMetrics = null;

// The current metric set
var curMetricSet = null;

// Dimentions for the results table longitudinal plot tooltip.
var resultsTooltipHeight = 220;
var resultsTooltipWidth = 400;

// Call handler to update window size
var winHeight = $(window).height();
var winWidth = $(window).width();
$(window).resize(function() {
    winHeight = $(window).height();
    winWidth = $(window).width();
});

// Poll the server to check for changes in the server state.
function pollServer() {
    if (polling === false) {
        polling = true;
        $.post("http://" + serverInfo.serverHost + ":" +
               serverInfo.serverPort + "/mobility_clinic/",
               {
                   requestType: "mobilityclinic",
                   type: "status",
               },
               function(d) {
                   serverStatus = d;
                   console.log(serverStatus);
                   polling = false;
                   if (serverStatus.clinicViewState === "TEST_DOCK_ERROR" &&
                       errorPageOpen === false) {
                       loadErrorPage(errorType.DOCK_ERROR);
                   } else if (serverStatus.clinicViewState === "TEST_CONFIGURATION_ERROR" &&
                       errorPageOpen === false) {
                           loadErrorPage(errorType.CONFIGURATION_ERROR);
                   } else if (serverStatus.clinicViewState === "TEST_FILE_ERROR" &&
                              errorPageOpen === false) {
                       loadErrorPage(errorType.FILE_ERROR);
                   } else if (serverStatus.clinicViewState === "TEST_CANNNOT_SYNC_RECORD_HEAD" &&
                              errorPageOpen === false) {
                       loadErrorPage(errorType.SYNC_ERROR);
                   }
               },
               "json");
    }
}

// Display an error page, and then wait until the server is ready to
// start a new test.
function loadErrorPage(error) {
    errorPageOpen = true;
    if (error === errorType.CONFIGURATION_ERROR) {
        postUIState(CONFIGURATION_ERROR_PAGE);
        $('#switchable-div').load('config_error.html', pageLoaded);
    } else if (error === errorType.DOCK_ERROR) {
        postUIState(DOCK_ERROR_PAGE);
        $('#switchable-div').load('error.html', pageLoaded);
    } else if (error === errorType.FILE_ERROR) {
        postUIState(FILE_ERROR_PAGE);
        $('#switchable-div').load('file_error.html', pageLoaded);
    } else if (error === errorType.SYNC_ERROR) {
        postUIState(SYNC_ERROR_PAGE);
        $('#switchable-div').load('sync_error.html', pageLoaded);
    } else {
        errorPageOpen = false;
        return;
    }

    function pageLoaded() {
        // Handle the "Next" button if this is a file error
        if (error === errorType.FILE_ERROR) {
            d3.select(".next_button").attr("disabled", "disabled")
                .on("click", loadStartPage);
        } else if (error === errorType.CONFIGURATION_ERROR) {
            attachConfigHandler();
        } else if (error === errorType.SYNC_ERROR) {
            d3.select(".next_button").on("click", loadStartPage);
        }
        // Go back to start page when the error clears
        var timer = setInterval(statusCheck, 500);
        function statusCheck() {
            if (serverStatus.clinicViewState === "TEST_WAITING_FOR_HW" ||
                serverStatus.clinicViewState === "TEST_READY_TO_START" ||
                serverStatus.clinicViewState === "TEST_WAITING_TO_RUN") {
                errorPageOpen = false;
                if (error === errorType.FILE_ERROR ||
                    error === errorType.SYNC_ERROR) {
                    clearInterval(timer);
                    d3.select(".next_button").attr("disabled", null);
                    return;
                }
                clearInterval(timer);
                loadStartPage();
            }
        }
    }
}

/**
 * Display the "Total Failure" error page. This is displayed when a
 * non-recoverable error is detected.
 * @todo Create the HTML page for the total failure.
 * @param errMsg Error message to display
 */
function loadTotalFaliure(errMsg) {
    console.warn(errMsg);
}


// Load the init page.
loadInitPage();

/**
 * Load the "Initializing" page and wait for server to finish loading
 * its data. Once initialization is done, ask if this is a streaming
 * record or not.
 */
function loadInitPage() {
    postUIState(WAITING_FOR_INIT);
    $('#switchable-div').load('initializing.html', pageLoaded);
    
    function pageLoaded() {
        var timer = setInterval(statusCheck, 500);
        
        function statusCheck() {
            if (serverStatus !== null &&
                serverStatus.clinicViewState !== "TEST_WAITING_FOR_INIT" &&
                serverStatus.clinicViewState !== "UNDEFINED") {
                clearInterval(timer);
                $.post("http://" + serverInfo.serverHost + ":" +
                       serverInfo.serverPort + "/mobility_clinic/",
                       {
                           requestType: "mobilityclinic",
                           type: "is_streaming",
                       },
                       function(d) {
                           isStreaming = d.isStreaming;
                           loadStartPage();
                       },
                       "json");
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}


/**
 * Load the launch page for Mobility Clinic and make sure the metric sets 
 * and definitions for the supported test types are loaded.
 * @todo: Only handles GaitTrack trials, so set the current metric set.
 * We'll change this up once we start handling balance trials as well.
 */
function loadStartPage() {
    postUIState(WAITING_FOR_DOCK);
    loadMetricSetsAndDefs();
    $('#switchable-div')
        .load('launch.html', pageLoaded);

    function pageLoaded() {
        d3.select("#start-gait-button")
            .on("click", function() {
                curMetricSet = gaitTrackMetrics;
                postTestType("IWalk");
                if (serverStatus !== null &&
                    (serverStatus.clinicViewState === "TEST_READY_TO_START" ||
                     serverStatus.clinicViewState === "TEST_WAITING_TO_RUN")) {
                    if (isStreaming) {
                        loadStreamAttachPage();
                    } else {
                        attachSensorsPage();
                    }
                } else if (serverStatus === null ||
                           serverStatus.clinicViewState === "TEST_WAITING_FOR_HW") {
                    if (isStreaming) {
                        loadStreamUndockPage();
                    } else {
                        redockPage();
                    }
                }
            });

        var balanceBtn = d3.select("#start-balance-button");
        if (isStreaming === false) {
            balanceBtn.attr("disabled", "disabled");
        } else {
            balanceBtn.on("click", function() {
                curMetricSet = iswayMetrics;
                postTestType("ISway");
                if (serverStatus !== null &&
                    (serverStatus.clinicViewState === "TEST_READY_TO_START" ||
                     serverStatus.clinicViewState === "TEST_WAITING_TO_RUN")) {
                    loadStreamAttachPage();
                } else if (serverStatus === null ||
                           serverStatus.clinicViewState === "TEST_WAITING_FOR_HW") {
                    loadStreamUndockPage();
                }
            });
        }

        attachConfigHandler();
        attachPowerHandlers();
        // TODO: Test code for viewing plots
        if (PLOT_DEBUG === true) {
            d3.select(".launch_buttons")
                .append("button").attr("class", "action_button large_button")
                .text("Plot test")
                .on("click", plotTest);
        }
    }
}

/**
 * Post the test type to the server
 */
function postTestType(testType) {
       $.post("http://" + serverInfo.serverHost + ":" +
              serverInfo.serverPort + "/mobility_clinic/",
              {
                  requestType: "mobilityclinic",
                  type: "test_type",
                  testType: testType,
              });
}


/**
 * Load the metric sets and the metric definitions for the supported trial types.
 */
function loadMetricSetsAndDefs() {
   d3.json("metric_sets/gait_track.json", function(err, metricSet) {
        if (err) {
            loadTotalFaliure(err);
            return;
        }
       // GaitTrack metric definitions
       $.post("http://" + serverInfo.serverHost + ":" +
              serverInfo.serverPort + "/mobility_clinic/",
              {
                  requestType: "mobilityclinic",
                  type: "get_metric_definitions",
                  testName: "IWalk",
              },
              function(d) {
                  gaitTrackMetrics = combineMetricsSetAndDefs(metricSet.GaitTrack, d);
              },
              "json");
       // ISway metric definitions
       $.post("http://" + serverInfo.serverHost + ":" +
              serverInfo.serverPort + "/mobility_clinic/",
              {
                  requestType: "mobilityclinic",
                  type: "get_metric_definitions",
                  testName: "ISway",
              },
              function(d) {
                  iswayMetrics = combineMetricsSetAndDefs(metricSet.ISway, d);
              },
              "json");

   });
}


/**
 * Attach handler to the "Configure" button
 */
function attachConfigHandler() {
    d3.select("#configure-button")
        .on("click", function() {
            $.post("http://" + serverInfo.serverHost + ":" +
                   serverInfo.serverPort + "/mobility_clinic/",
                   {
                       requestType: "mobilityclinic",
                       type: "configure_monitors",
                   }
                  );
        });
}

/**
 * Attach handlers to power off, and power on buttons.
 */
function attachPowerHandlers() {

    d3.select("#power-off-button")
        .on("click", function() {
            $.post("http://" + serverInfo.serverHost + ":" +
                   serverInfo.serverPort + "/mobility_clinic/",
                   {
                       requestType: "mobilityclinic",
                       type: "power_off_monitors",
                   }
                  );
        });

    d3.select("#power-on-button")
        .on("click", function() {
            $.post("http://" + serverInfo.serverHost + ":" +
                   serverInfo.serverPort + "/mobility_clinic/",
                   {
                       requestType: "mobilityclinic",
                       type: "power_on_monitors",
                   }
                  );
        });
}


/**
 * Display the "Redock monitors" page. Calls loading of the "Attach
 * sensors page" when docked monitors are detected by server.
 */
function redockPage() {
    $('#switchable-div')
        .load('dock.html', pageLoaded);

    function pageLoaded() {
        back = loadStartPage;
        next = attachSensorsPage;
        navButtonHandlers();

        var timer = setInterval(statusCheck, 500);
        
        function statusCheck() {
            if (serverStatus.clinicViewState === "TEST_READY_TO_START" ||
                serverStatus.clinicViewState === "TEST_WAITING_TO_RUN") {
                clearInterval(timer);
                attachSensorsPage();
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}

function attachSensorsPage() {
    $('#switchable-div')
        .load('attach.html',
              pageLoaded);

    function pageLoaded() {
        var timer = setInterval(statusCheck, 500);
        back = function() {
            clearInterval(timer);
            postUIState(RESTARTING_TEST);
            loadStartPage();
        };
        next = function() {
            clearInterval(timer);
            walkPage();
        };
        navButtonHandlers();

        postUIState(WAITING_FOR_ATTACH);
        d3.select(".next_button").attr("disabled", "disabled");
        
        function statusCheck() {
            if (serverStatus.clinicViewState === "TEST_RUNNING") {
                d3.select(".next_button").attr("disabled", null);
            } else if (serverStatus.clinicViewState === "TEST_WAITING_FOR_HW") {
                clearInterval(timer);
                redockPage();
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}

var seconds = 120;
var countdownTimer = null;
function walkPage() {
    console.log("walkPage");
    $('#switchable-div').load('walk.html', pageLoaded);

    function pageLoaded() {
        back = attachSensorsPage;
        next = dockSensorsPage;
        navButtonHandlers();
        postUIState(RUNNING_TEST);
        // Handle the timer button
        var timerButton = d3.select("#time-button");
        var timerText = d3.select(".countdown_timer");
        var seconds = 120;
        var timerStarted = false;

        timerButton.on("click", function() {
            if (!timerStarted) {
                countdownTimer = setInterval(updateTimer, 1000);
                d3.select(this).text("Reset");
                timerStarted = true;
            } else if (errorPageOpen === true) {
                clearInterval(countdownTimer);
                countdownTimer = null;
            } else {
                clearInterval(countdownTimer);
                countdownTimer = null;
                d3.select(this).text("Start");
                timerText.style("color", "black").text("120");
                seconds = 120;
                timerStarted = false;
            }
        });

        function updateTimer() {
            seconds--;
            timerText.text(seconds);
            if (seconds === 0) {
                timerText.style("color", "red");
                clearInterval(countdownTimer);
            }
        }
    }
}

function dockSensorsPage() {
    console.log("dockSensorsPage");
    $('#switchable-div').load('dock.html', pageLoaded);

    // Kill the countdown timer if it's still running
    if (countdownTimer !== null) {
        clearInterval(countdownTimer);
        seconds = 120;
    }

    postUIState(WAITING_FOR_DOCK);
    function pageLoaded() {
        // Move on if the sensors are docked
        if (serverStatus.clinicViewState === "TEST_PROCESSING" ||
            serverStatus.clinicViewState === "TEST_COMPLETE" ||
            serverStatus.clinicViewState === "TEST_COPYING_FILES" ||
            serverStatus.clinicViewState === "TEST_CONVERTING_FILES") {
            analyzeResultsPage();
            return;
        }

        var timer = setInterval(statusCheck, 500);
        
        function statusCheck() {
            if (serverStatus.clinicViewState === "TEST_PROCESSING" ||
                serverStatus.clinicViewState === "TEST_COMPLETE" ||
                serverStatus.clinicViewState === "TEST_COPYING_FILES" ||
                serverStatus.clinicViewState === "TEST_CONVERTING_FILES") {
                clearInterval(timer);
                analyzeResultsPage();
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}

/**
 * Display the "Analyzing results page" and wait for the trial to
 * finish analyzing. Once finished, get the study data and the metric
 * definitions for the test, then open the reports page.
 * @TODO: Defaults to just getting the metric definitions for GaitTrack.
 *  Should determine which definitions to get based on the current trial type.
 */

function analyzeResultsPage() {
    console.log("analyzeResultsPage");
    $('#switchable-div').load('analyze.html', pageLoaded);
    
    postUIState(WAITING_FOR_ANALYSIS);

    function pageLoaded() {
        var timer = setInterval(statusCheck, 500);

        function statusCheck() {
            if (serverStatus === null) {
                return;
            }
            var infoText = d3.select(".instruction_big").select("p");
            if (serverStatus.clinicViewState === "TEST_COPYING_FILES") {
                infoText.text("Copying data from monitors.");
            } else if (serverStatus.clinicViewState === "TEST_CONVERTING_FILES") {
                infoText.text("Converting monitor data.");
            } else if (serverStatus.clinicViewState === "TEST_PROCESSING") {
                infoText.text("Analyzing monitor data.");
            } else if (serverStatus.clinicViewState === "TEST_PROCESSED") {
                clearInterval(timer);
                $.post("http://" + serverInfo.serverHost + ":" +
                       serverInfo.serverPort + "/mobility_clinic/",
                       {
                           requestType: "mobilityclinic",
                           type: "get_study",
                       },
                       function(d) {
                           reportPage(d, UNASSIGNED_SUBJECT);
                       },
                       "json");
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}

/**
 * Display the report page for a trial.
 * @param studyData The data associated with the Mobility Clinic study.
 * @param publicId The public ID of the subject associated with the trial.
 * @param trial The trial to display. Defaults to most recent for the
 *        subject if none is given.
 */
function reportPage(studyData, publicId, trial) {
    console.log("reportPage");
    $('#switchable-div').load('report_table.html', pageLoaded);

    postUIState(READING_RESULTS);

    function pageLoaded() {
        next = loadStartPage;
        navButtonHandlers();

        d3.select(".add_subject_button")
            .on("click", function() { return addSubjectsPage(studyData, publicId, trial); });
        d3.select(".view_history_button").attr("disabled", null)
            .on("click", function() {
                return showHistoryPage(studyData, publicId);
            });
        if (!publicId || publicId === UNASSIGNED_SUBJECT) {
            d3.select(".view_history_button")
                .attr("disabled", "disabled");
        }

        buildSubjectSelect(studyData, publicId);
        if (trial) {
            buildResultsTable(studyData, trial, publicId, curMetricSet.testType,
                              curMetricSet.metrics);
        } else {
            buildResultsTable(studyData, findCurrentTrial(studyData, publicId),
                              publicId, curMetricSet.testType,
                              curMetricSet.metrics);
        }
    }
}

/**
 * Find the current trial for a given subject.
 * @param studyData The data associated with the Mobility Clinic study.
 * @param publicId The public ID of the subject to match.
 * @return The most recent trial for the subject.
 */
function findCurrentTrial(studyData, publicId) {
    if (!publicId) {
        publicId = UNASSIGNED_SUBJECT;
    }
    var trials = findAllTrialsForSubject(studyData, publicId);
    trials.sort(function (a, b) { return d3.descending(a.date, b.date); });
    return trials[0];
}

/**
 * Find trial matching a given date in a list of trials.
 * @param studyData The data associated with the Mobility Clinic study.
 * @param publicId The public ID of the subject to match.
 * @param date The date to match.
 * @return The trial that matches the date, or null if no match.
 */
function findTrialByDate(studyData, publicId, date) {
    var trials = findAllTrialsForSubject(studyData, publicId);
    for (var i = 0; i < trials.length; i++) {
        if(utils.getDateFromString(trials[i].date).getTime() ===
           date.getTime()) {
            return trials[i];
        }
    }
    return null;
}

/**
 * Get all the trials for a subject by publicId.
 * @param studyData The data associated with the Mobility Clinic study.
 * @param publicId The public ID of the subject.
 * @return An array of trials for the subject, or null if no match.
 */
function findAllTrialsForSubject(studyData, publicId) {
    var subject = findSubjectInStudy(studyData, publicId);
    if (subject !== null) {
        return subject.children[0].children;
    }
    return null;
}

/**
 * Find the subject in the study data that matches the given public ID.
 * @param studyData The Mobility Clinic data to search,
 * @param publicId The ID to match to subject.
 * @return The matching study subject, or null if no match
 */
function findSubjectInStudy(studyData, publicId) {
    for (var i = 0; i < studyData.children.length; i++) {
        if (studyData.children[i].publicId === publicId)
            return studyData.children[i];
    }
    return null;
}

/**
 * Build the drop down and buttons for the subject selection on the
 * report.html page.
 * @param studyData The Mobility Clinic data
 * @param pubicId The currently selected subject's public id.
 */
function buildSubjectSelect(studyData, publicId) {
    var select = d3.select("select")
        .on("change", function() { return subjectSelected(this); });
    select.selectAll("option")
        .data(studyData.children)
        .enter()
        .append("option")
        .each( function(d) {
            var opt = d3.select(this);
            opt.attr("value", d.publicId + ":" + d.id)
                .text(d.publicId);
            if (d.publicId === publicId)
                opt.attr("selected", "selected");
        });
}

/**
 * Display the results for the current trial. If possible, also
 * display the percent change from baseline and previous trials.
 * Change is not displayed if publicId == UNASSIGNED_SUBJECT.
 * Requires that the "report.html" page has been loaded.
 * @param studyData All data for the Mobility Clinic study.
 * @param currTrial The trial whose data to display.
 * @param publicId Public ID of the subject.
 * @param testType The type of test to display in table.
 * @param metricSet The metrics from the trial to be displayed .
 */
function buildResultsTable(studyData, currTrial, publicId, testType, metricSet) {
    // Get just the date metrics from the trial
    var trials = findAllTrialsForSubject(studyData, publicId);
    trials = getDateAndMetricsFromTrials(trials, metricSet);
    trials = filterTrialsByTestType(trials, testType);
    trials.sort(function (a, b) { return d3.descending(a.date, b.date); });
    trials.forEach(function(d) {
        d.metrics = d3.nest()
            .key(function(e) { return e.metricName; })
            .map(d.metrics, d3.map);
    });
    currTrial = getDateAndMetricsFromTrials([currTrial], metricSet)[0];
    currTrial.metrics = d3.nest()
        .key(function(d) { return d.metricName; })
        .map(currTrial.metrics, d3.map);

    // Only do the first trial if the subject has not been assigned
    if (publicId === UNASSIGNED_SUBJECT) {
        trials = [trials[0]];
    }

    if (trials.length === 1) {
        d3.select(".report_table_div").style("width", "60%%");
    }

    buildLeftResultsTable(trials, currTrial, metricSet);
    buildCenterResultsTable(trials, metricSet);
    if (publicId !== UNASSIGNED_SUBJECT && trials.length > 1) {
        buildRightResultsTable(trials, metricSet, true);
    } else {
        buildRightResultsTable(trials, metricSet, false);
    }
}

/**
 * Build the left results table which holds metric names and the
 * results of the most recent trial.
 * @param trials Array of all trials for subject.
 * @param currTrial The current trial to display.
 * @param metricSet The metrics to display.
 */
function buildLeftResultsTable(trials, currTrial, metricSet) {
    var headers = ["Metric",
                   utils.getShortFormattedDate(currTrial.date)];
    var leftTable = d3.select(".left_table");
    var thead = leftTable.append("thead");
    var tbody = leftTable.append("tbody");

    var theadRows = thead.selectAll("tr")
        .data([headers])
        .enter()
        .append("tr").attr("class", "reports_head_row");
    theadRows.selectAll("th")
        .data(function(d) { return d; })
        .enter()
        .append("th").text(function(d) { return d; });

    // Build a matrix of metric names and baseline values, if we have a baseline
    var matrix = [];
    for (var prop in metricSet) {
        var metric = metricSet[prop].name;
        var entry = [];
        entry.push({ metricSetProp: prop,
                      displayValue: metricSet[prop].displayName +
                                    " (" + metricSet[prop].units + ")"});
        var value = { metricSetProp: prop,
                  displayValue: "--" };
        if (currTrial.metrics[metric]) {
            var metricValue = currTrial.metrics[metric][0].mean;
            if (isNaN(metricValue) !== true) {
                value.displayValue = metricValue.toPrecision(3);
            }
        }
        entry.push(value);
        matrix.push(entry);
    }
    
    var tbodyRows = tbody.selectAll("tr")
        .data(matrix)
        .enter()
        .append("tr")
        .each(function(d, i) {
            if (i % 2 !== 0) {
                d3.select(this).attr("class", "reports_odd_row");
            }
            if (trials.length > 1) {
                d3.select(this)
                    .on("mouseover",  function () {
                        return showLongitudinalTooltip(this.__data__, trials, metricSet);
                    })
                    .on("mousemove", updateLongitudinalTooltip)
                    .on("mouseout", hideLongitudinalTooltip);
            }
        });
    tbodyRows.selectAll("td")
        .data(function(d) { return d; })
        .enter()
        .append("td")
        .each(function(d,i) {
            var td = d3.select(this);
            td.text(d.displayValue);
            if (i % 2 !== 0) {
                td.attr("class", "center_cell");
            } else {
                td.attr("class", "name_cell");
            }
        });
}

/**
 * Build the center results table which contains scrolling view of all
 * previous trials. Don't display anything if the array of trials only
 * has one element, since that should be displayed in the right table.
 * @param trials The array of trials for the subject.
 * @param metricSet The set of trials we care about.
 */
function buildCenterResultsTable(trials, metricSet) {
    if (trials.length <= 1) {
        d3.select(".center_table_div").remove();
        return;
    }

    var displayTrials = trials.slice(1);

    // Build matrix for central table history stuff
    var dataMatrix = [];
    var k = 0;
    for (var prop in metricSet) {
        for (var i = 0; i < displayTrials.length; i++) {
            var value = { metricSetProp: prop, displayValue: "--" };
            if (displayTrials[i].metrics[metricSet[prop].name]) {
                var metricValue = displayTrials[i]
                    .metrics[metricSet[prop].name][0].mean;
                if (isNaN(metricValue) === false) {
                    value.displayValue = metricValue.toPrecision(3);
                }
            }
            if (dataMatrix[k]) {
                dataMatrix[k].push(value);
            } else {
                dataMatrix[k] = [value];
            }
        }
        k++;
    }

    var centerTbl = d3.select(".center_table");
    var theadRows = centerTbl.append("thead").append("tr");
    theadRows.selectAll("th")
        .data(displayTrials)
        .enter()
        .append("th").attr("class", "center_cell reports_head_row")
        .each(function(d, i) {
            var th = d3.select(this);
            th.text(utils.getShortFormattedDate(d.date));
            // Create a reference to the last cell, table canget
            // scrolled to the most recent trial
            if (i === displayTrials.length - 1) {
                th.attr("id", "latest-trial");
            }
        });

    var tbody = centerTbl.append("tbody");
    var tbodyRows = tbody.selectAll("tr")
        .data(dataMatrix)
        .enter()
        .append("tr")
        .each(function(d, i) {
            if (i % 2 !== 0) {
                d3.select(this).attr("class", "reports_odd_row");
            }
            if (trials.length > 1) {
                d3.select(this)
                    .on("mouseover",  function () {
                        return showLongitudinalTooltip(this.__data__, trials, metricSet);
                    })
                    .on("mousemove", updateLongitudinalTooltip)
                    .on("mouseout", hideLongitudinalTooltip);
            }
        });

    tbodyRows.selectAll("td")
        .data(function(d) { return d; })
        .enter()
        .append("td").attr("class", "center_cell")
        .text(function(d) { return d.displayValue; });
}

/**
 * Build the right most trial results table with "Normative", "Goal",
 * and "Baseline" values.
 * @param trials The set of trials being displayed.
 * @param metricSet The metrics to display.
 * @param baseline Flag to determine if "Baseline" values should be displayed.
 */
function buildRightResultsTable(trials, metricSet, baseline) {
    var headers = [];
    if (baseline === true) {
        headers.push("Baseline " +
                     utils.getShortFormattedDate(trials[trials.length - 1].date));
    }
    headers = headers.concat(["Normative", "Goal"]);
    var rightTbl = d3.select(".right_table");
    
    var theadRows = rightTbl.append("thead")
        .append("tr").attr("class", "reports_head_row");
    theadRows.selectAll("th")
        .data(headers)
        .enter()
        .append("th")
        .text(function(d) { return d; });

    var matrix = [];
    for (var prop in metricSet) {
        var baselineVal = "--";
        var normVal = "--";
        var goalVal = "--";

        if (metricSet[prop].meanNorm) {
            normVal = metricSet[prop].meanNorm.toPrecision(3);
        }
       if (baseline === false) {
            matrix.push([{metricSetProp: prop, displayValue: normVal},
                         {metricSetProp: prop, displayValue: goalVal}]);
        } else {
            if (trials[trials.length - 1].metrics[metricSet[prop].name]) {
                baselineVal = trials[trials.length - 1]
                              .metrics[metricSet[prop].name][0].mean
                              .toPrecision(3);
            }
            matrix.push([{metricSetProp: prop, displayValue: baselineVal},
                         {metricSetProp: prop, displayValue: normVal},
                         {metricSetProp: prop, displayValue: goalVal}]);
        }
    }
    var tbody = rightTbl.append("tbody");
    var tbodyRows = tbody.selectAll("tr")
        .data(matrix)
        .enter()
        .append("tr")
        .each(function(d, i) {
            if (i % 2 !== 0) {
                d3.select(this).attr("class", "reports_odd_row");
            }
            if (trials.length > 1) {
                d3.select(this)
                    .on("mouseover",  function () {
                        return showLongitudinalTooltip(this.__data__, trials, metricSet);
                    })
                    .on("mousemove", updateLongitudinalTooltip)
                    .on("mouseout", hideLongitudinalTooltip);
            }
        });

    tbodyRows.selectAll("td")
        .data(function(d) { return d; })
        .enter()
        .append("td")
        .each(function(d, i) {
            var td = d3.select(this);
            if (baseline === true && (i % 3 === 0)) {
                td.attr("class", "baseline_cell");
            } else {
                td.attr("class", "center_cell");
            }
            td.text(d.displayValue);
        });
}


/**
 * Handler called when subject is selected from "select" dropdown.
 * Assigns the current trial being displayed to the selected subject,
 * and enables the "View History" option if subject is not
 * UNASSIGNED_SUBJECT.
 * @param option The option selected in dropdown.
 */
function subjectSelected(option) {
    console.log("subject selected: " + this.value);
    var values = option.value.split(":");
    $.post("http://" + serverInfo.serverHost + ":" +
           serverInfo.serverPort + "/mobility_clinic/",
           {
               requestType: "mobilityclinic",
               type: "change_subject",
               publicId: values[0],
               id: values[1],
           },
           function(d) {
               // Handle "View History" button
               var btn = d3.select(".view_history_button");
               if (values[0] !== UNASSIGNED_SUBJECT) {
                   btn.attr("disabled", null)
                       .on("click", function() {
                           return showHistoryPage(d, values[0]);
                       });
               } else {
                   btn.attr("disabled", "disabled");
               }
               reportPage(d, values[0]);
           },
           "json");
}

/**
 * Add a new subject to the Mobility Clinic study.
 * @param studyData The Mobility Clinic data.
 * @param publicId The id of subject that was previously on the
 * results page.
 * @param currTrial The trial that was being displayed on the results table.
 */
function addSubjectsPage(studyData, publicId, currTrial) {
    console.log("addSubjectsPage");
    $('#switchable-div').load('add_subject.html', pageLoaded);

    function pageLoaded() {
        back = function() { return reportPage(studyData, publicId, currTrial); };
        navButtonHandlers();

        d3.select(".save_button").on("click", saveSubject);
    }

    function saveSubject() {
        var publicId = d3.select("#public-id").node().value;
        var yearOfBirth = d3.select("#year-of-birth").node().value;
        var height = d3.select("#height").node().value;
        var control = d3.select("#control").node().checked;

        $.post("http://" + serverInfo.serverHost + ":" +
               serverInfo.serverPort + "/mobility_clinic/",
               {
                   requestType: "mobilityclinic",
                   type: "new_subject",
                   publicId: publicId,
                   yearOfBirth: yearOfBirth,
                   height: height,
                   control: control,
               },
               function(d) {
                   reportPage(d, publicId);
               },
               "json");
    }
}

/**
 * Post the given state to the server.
 * @param state The state to post.
 */
function postUIState(state) {
    console.log("Posting state: " + state);
    $.post("http://" + serverInfo.serverHost + ":" +
           serverInfo.serverPort + "/mobility_clinic/",
           {
               requestType: "mobilityclinic",
               type: "ui_state",
               uiState: state,
           });
}

function navButtonHandlers() {
    d3.select(".back_button").on("click", back);
    d3.select(".next_button").on("click", next);
}

/**
 * Display the history page for the subject with the given public ID.
 * Shows the longitudinal plots for all the metrics that are displayed in
 * on the results page.
 * @param studyData The data for this Mobility Clinic study.
 * @param publicId The ID of the subject.
 */
function showHistoryPage(studyData, publicId) {
    console.log("showHistoryPage");
    $('#switchable-div').load('view_history.html', pageLoaded);

    function pageLoaded() {
        next = loadStartPage;
        back = function() { return reportPage(studyData, publicId); };
        navButtonHandlers();

        // Set name in title
        d3.select("#subject-publicid").text(publicId);

        // Get the subject trials, then send them to plot.
        var trials = findAllTrialsForSubject(studyData, publicId);
        // TODO: Determine if we should be passing the test type into
        // this function instead of getting it from the current metric
        // set.
        var plotData = formatLongitudinalPlotData(trials, curMetricSet.testType,
                                                  curMetricSet.metrics);

        // Build plots for each metric.
        var div = d3.select(".history_plots");
        plotData.forEach(function(d) {
            var plot = longitudinalPlot().width(800).height(200)
                .margin({top:40, bottom:40, left:80, right:80})
                .x(function(d) { return d.date; })
                .y(function(d) { return d.value; })
                .extra(function(d) { return d.change; })
                .zoomEnabled(true)
                .tooltipsEnabled(true)
                .tooltipHtml(
                    function(d) {
                        var baseChange = "--";
                        var prevChange = "--";
                        if (d[2] !== null) {
                            if (d[2].fromBaseline !== null) {
                                baseChange = d[2].fromBaseline.toPrecision(3);
                            }
                            if (d[2].fromPrev !== null) {
                                prevChange = d[2].fromPrev.toPrecision(3);
                            }
                        }
                        return "<strong>Date: </strong>" +
                            utils.getStringFromDate(d[0]) +
                            "<br><strong>Value: </strong>" +
                            d[1].toPrecision(3) +
                            "<br><strong>Change from baseline: </strong>" +
                            baseChange + "%" +
                            "<br><strong>Change from previous: </strong>" +
                            prevChange + "%";
                    })
                .pointOnClickHandler(
                    function(d) {
                        console.log("Click on point");
                        return displayReportForTrial(studyData, publicId, d[0]);
                    });
            if (d.minPlotY !== undefined && d.minPlotY !== null) {
                plot.minY(d.minPlotY);
            }
            if (d.meanNorm) {
                plot.normMean(d.meanNorm);
            }
            var plotDiv = div.append("div").attr("id", "longitudinal-plot");
            plot.xAxisLabel("Time/Date").title(d.name);
            plotDiv.datum(d.values).call(plot);
        });
    }
}

/**
 * Build an array where each entry in the array is an object
 * containing all the data across trials for a given metric.
 * @param trials The list of trials whose data we are to format.
 * @param testType The test type of the trials to display.
 * @param metricSet The metrics to pull out of the trials.
 */
function formatLongitudinalPlotData(trials, testType, metricSet) {
    var plotData = [];
    trials = getDateAndMetricsFromTrials(trials, metricSet);
    trials = filterTrialsByTestType(trials, testType);
    trials.sort(function (a, b) { return d3.ascending(a.date, b.date); });
    trials.forEach(function(d) {
        d.metrics = d3.nest()
            .key(function(e) { return e.metricName; })
            .map(d.metrics, d3.map);
    });
    var firstTrial = trials[0];

    for (var prop in metricSet) {
        var metricName = metricSet[prop].name;
        var metricData = { name: metricSet[prop].displayName,
                           meanNorm: metricSet[prop].meanNorm,
                           values: [] };
        var base = null;
        if (metricSet[prop].minPlotY !== undefined &&
            metricSet[prop].minPlotY !== null) {
            metricData.minPlotY = metricSet[prop].minPlotY;
        }
        if (firstTrial.metrics[metricName]) {
            base = firstTrial.metrics[metricName][0].mean;
        }

        for (var i = 0; i < trials.length; i++) {
            if (trials[i].metrics[metricName]) {
                var value = trials[i].metrics[metricName][0].mean;
                if (isNaN(value) === true) {
                    continue;
                }
                var fromBaseline = null;
                if (base !== null) {
                    fromBaseline = utils.percentChange(value, base);
                }

                var fromPrev = null;
                if (i > 0 && trials[i - 1].metrics[metricName]) {
                    fromPrev = utils.percentChange(value,
                                                   trials[i - 1]
                                                   .metrics[metricName][0]
                                                   .mean);
                }
                                                  
                var change = { fromBaseline: fromBaseline,
                               fromPrev: fromPrev };
                metricData.values.push({ value: value,
                                         date: utils.getDateFromString(trials[i].date),
                                         change: change });
            }
        }
        plotData.push(metricData);
    }
    return plotData;
}


/**
 * Given an array of trials, return an array of objects containing the
 * trial date, the trial test type, and an array of just the metrics
 *  we care about.
 * @param trials List of trials to pull metrics from.
 * @param metricSet The metrics that we want from the trials.
 * @return Any array of truncated trial objects.
 */
function getDateAndMetricsFromTrials(trials, metricSet) {
    var trialData = trials.map(function(d) {
        return { date: d.date,
                 testType: d.testDefinition.testName,
                 metrics: getMetricSetMetrics(d.metrics, metricSet) };
    });
    return trialData;
}

/**
 * Given a list of trials generated by a call to getDateAndMetricsFromTrials,
 * filter out all trials that do not match the given test type.
 * @param trials The list of trials to filter.
 * @param testType The test type to filter with.
 */
function filterTrialsByTestType(trials, testType) {
    var trialData = [];
    trials.forEach(function(d) {
        if (d.testType === testType) {
            trialData.push(d);
        }
    });
    return trialData;
}

/**
 * Push the needed fields from the metric definitions into the metric
 * set so they can be accessed later.
 * @param metricSet The object containing the metrics that will be handled.
 * @param defs The complete set of metric definitions.
 * @retrun The metric set with definition data.
 */
function combineMetricsSetAndDefs(metricSet, defs) {
    defs = d3.nest()
        .key(function(d) { return d.metricName; })
        .map(defs, d3.map);

    var newSet = {testType: metricSet.testType,
                  conditionName: metricSet.conditionName,
                  metrics: {}};
    var metricList = metricSet.metrics;
    for (var i = 0; i < metricList.length; i++) {

        if (metricList[i].charAt(metricList[i].length - 1) === "L" &&
            defs[metricList[i].substring(0, metricList[i].length - 2)]) {
            newSet.metrics[metricList[i]] = {};
            newSet.metrics[metricList[i]].name = metricList[i];
            newSet.metrics[metricList[i]].displayName = metricList[i];
            newSet.metrics[metricList[i]].meanNorm =
                defs[metricList[i].substring(0, metricList[i].length - 2)][0].meanNormL;
            newSet.metrics[metricList[i]].group =
                defs[metricList[i].substring(0, metricList[i].length - 2)][0].metricGroup;
            newSet.metrics[metricList[i]].units =
                defs[metricList[i].substring(0, metricList[i].length - 2)][0].units;
        } else if (metricList[i].charAt(metricList[i].length - 1) === "R" &&
                   defs[metricList[i].substring(0, metricList[i].length - 2)]) {
            newSet.metrics[metricList[i]] = {};
            newSet.metrics[metricList[i]].name = metricList[i];
            newSet.metrics[metricList[i]].displayName = metricList[i];
            newSet.metrics[metricList[i]].meanNorm =
                defs[metricList[i].substring(0, metricList[i].length - 2)][0].meanNormR;
            newSet.metrics[metricList[i]].group =
                defs[metricList[i].substring(0, metricList[i].length - 2)][0].metricGroup;
            newSet.metrics[metricList[i]].units =
                defs[metricList[i].substring(0, metricList[i].length - 2)][0].units;
        } else if (defs[metricList[i]]) {
            newSet.metrics[metricList[i]] = {};
            newSet.metrics[metricList[i]].name = metricList[i];
            newSet.metrics[metricList[i]].displayName = metricList[i];
            newSet.metrics[metricList[i]].meanNorm =
                defs[metricList[i]][0].meanNorm;
            newSet.metrics[metricList[i]].group =
                defs[metricList[i]][0].metricGroup;
            newSet.metrics[metricList[i]].units =
                defs[metricList[i]][0].units;
        }
    }
    return newSet;
}

/**
 * Pull just the metrics we care about out of an array of trial metrics.
 * @param metrics List of metrics to filter.
 * @param metricSet The metric set containing the metrics we want.
 */
function getMetricSetMetrics(metrics, metricSet) {
    var results = [];
    metrics.forEach(function(d) {
        for (var prop in metricSet) {
            if (metricSet[prop].name === d.metricName) {
                results.push(d);
            }
        }
    });
    return results;
}

/**
 * Display the report for a given subject and trial date from the study data
 * @param studyData The data associated with the Mobility Clinic study
 * @param publicId The public ID of the subject
 * @param trialDate The date of the trial
 */
function displayReportForTrial(studyData, publicId, trialDate) {
    var trial = findTrialByDate(studyData, publicId, trialDate);
    reportPage(studyData, publicId, trial);
}


// TODO: Testing function for generating longitudinal plots without
// having to run a trial
function plotTest() {
    curMetricSet = gaitTrackMetrics;
    $.post("http://" + serverInfo.serverHost + ":" +
           serverInfo.serverPort + "/mobility_clinic/",
           {
               requestType: "mobilityclinic",
               type: "get_study",
           },
           function(d) {
               showHistoryPage(d, "jess"); // Assume the subject's publicId
           },
           "json");
}

var tooltip = d3.select("body").append("div")
	.attr("class", "plot_tooltip").text("");

/**
 * Create the results table longitudinal plot tooltip and make it visible.
 * @param rowData The data contained in the row being moused on.
 * @param trials Array of all trial data for current subject.
 * @param metricSet The metrics being displayed in the table.
 */
function showLongitudinalTooltip(rowData, trials, metricSet) {
    var plotDiv = tooltip.append("div").attr("id", "results-plot-tooltip");
    var plotData = [];
    trials.forEach( function(d) {
        if(d.metrics[metricSet[rowData[0].metricSetProp].name]) {
            var value = d.metrics[metricSet[rowData[0].metricSetProp].name][0].mean;
            if (isNaN(value)) {
                return;
            }
            plotData.push({
                date: utils.getDateFromString(d.date),
                value: value.toPrecision(3)
            });
        }
    });

    var plot = longitudinalPlot()
        .width(resultsTooltipWidth).height(resultsTooltipHeight)
        .margin({top:20, bottom:40, left:40, right:10})
        .x(function(d) { return d.date; })
        .y(function(d) { return d.value; })
        .xAxisLabel("Time/Date")
        .title(metricSet[rowData[0].metricSetProp].displayName);
    if (metricSet[rowData[0].metricSetProp].minPlotY !== undefined &&
        metricSet[rowData[0].metricSetProp].minPlotY !== null) {
        plot.minY(metricSet[rowData[0].metricSetProp].minPlotY);
    }

    plotDiv.datum(plotData).call(plot);

    updateLongitudinalTooltip();

    tooltip.style("visibility", "visible");
}

/**
 * Update the position of the results table tooltip based on position
 * of the mouse.
 */
function updateLongitudinalTooltip() {
    if (d3.event !== undefined) {
        var yOffset = 10;
        if ((d3.event.clientY + resultsTooltipHeight - yOffset/2) > winHeight) {
            yOffset = (d3.event.clientY + resultsTooltipHeight + yOffset/2) - winHeight;
        }
        tooltip.style("top", (d3.event.pageY - yOffset) + "px")
            .style("left", (d3.event.pageX + 10) + "px");
    }
}

/**
 * Hide the results table longitudial plot tooltip div and remove the
 * plot.
 */
function hideLongitudinalTooltip() {
    tooltip.style("visibility", "hidden");
    d3.select("#results-plot-tooltip").remove();
}

//
//
// Section that handles UI flow for streaming trials
//
// TODO: Consider moving this into another file, or moving the shared
// reporting code so that this file doesn't become to unwieldly.
//

/**
 * Load the streaming "Undock monitors" page.
 */
function loadStreamUndockPage() {
    postUIState(WAITING_FOR_HW);
    $('#switchable-div')
        .load('stream_undock.html', pageLoaded);

    function pageLoaded() {
        back = loadStartPage;
        next = attachSensorsPage;
        navButtonHandlers();

        var timer = setInterval(statusCheck, 500);
        
        function statusCheck() {
            if (serverStatus.clinicViewState === "TEST_WAITING_TO_RUN") {
                clearInterval(timer);
                $.post("http://" + serverInfo.serverHost + ":" +
                       serverInfo.serverPort + "/mobility_clinic/",
                       {
                           requestType: "mobilityclinic",
                           type: "get_test_definition",
                           testType: curMetricSet.testType,
                           conditionName: curMetricSet.conditionName,
                       },
                       function(d) {
                           loadStreamRunTrialPage(d);
                       },
                       "json");
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}

/**
 * Load the run trial page for a streaming test. Use the given test
 * test definition to determine what to use for timers.
 * @param testDef The test definition for the trial begin run.
 */
function loadStreamRunTrialPage(testDef) {
    $('#switchable-div')
        .load('stream_run_trial.html', pageLoaded);

    function pageLoaded() {
        console.log(testDef);
        // Set up the run trial page according to the test definition
        d3.select(".subject_instructions").text(testDef.subjectInstructions);
        var timer = trialTimer.init(testDef, d3.select(".trial_timer_value"));
        var status = null;
        d3.select("#start-trial-button")
            .on("click", function() {
                postUIState(RUNNING_TEST);
                var btn = d3.select(this);
                timer.start();
                status = setInterval(checkTimer, 500);
                btn.text("Stop")
                    .on("click", function() {
                        timer.stopMain();
                    });
            });
        
        function checkTimer() {
            if (timer.isDone() === true) {
                clearInterval(status);
                postUIState(WAITING_FOR_ANALYSIS);
                loadStreamBufferPage();
            }
        }
    }
}


function loadStreamBufferPage() {
    console.log("analyzeResultsPage");
    $('#switchable-div').load('analyze.html', pageLoaded);
    
    postUIState(WAITING_FOR_ANALYSIS);

    function pageLoaded() {
        var timer = setInterval(statusCheck, 500);
        d3.select(".instruction_big").select("p")
            .text("Buffering data from  monitors.");

        function statusCheck() {
            if (serverStatus === null) {
                return;
            }
            var infoText = d3.select(".instruction_big").select("p");
            if (serverStatus.clinicViewState === "TEST_PROCESSING") {
                infoText.text("Analyzing monitor data.");
            } else if (serverStatus.clinicViewState === "TEST_PROCESSED") {
                clearInterval(timer);
                $.post("http://" + serverInfo.serverHost + ":" +
                       serverInfo.serverPort + "/mobility_clinic/",
                       {
                           requestType: "mobilityclinic",
                           type: "get_study",
                       },
                       function(d) {
                           reportPage(d, UNASSIGNED_SUBJECT);
                       },
                       "json");
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}


/**
 * Module that handles the trial timer based on the values taken from a
 * test definition.
 */
var trialTimer = (function () {
    var timer = {};
    var startDelay;
    var endDelay;
    var recordDuration;
    var recordType;
    var clock;
    var clockRunning;
    var runEndDelay;
    var runStartDelay;
    var runMainTimer;
    var isIndeterminate;
    var isCountdown;
    var isDone;
    var seconds;
    var timerText;
    var stopMain;

    timer.init = function(testDef, text) {
        startDelay = parseInt(testDef.startDelay, 10);
        endDelay = parseInt(testDef.endDelay, 10);
        recordDuration = parseInt(testDef.recordDuration, 10);
        recordType = testDef.recordType;
        timerText = text;
        clock = null;
        clockRunning = false;
        runEndDelay = false;
        runStartDelay = false;
        runMainTimer = true;
        isIndeterminate = false;
        isCountdown = false;
        isDone = false;
        stopMain = false;

        if (recordType === "INDETERMINATE") {
            isIndeterminate = true;
            seconds = 0;
        } else {
            isCountdown = true;
            seconds = recordDuration;
        }

        if (startDelay > 0) {
            runStartDelay = true;
            isCountdown = true;
            seconds = startDelay;
            timerText.style("color", "red");
        }
        timerText.text(seconds);

        if (endDelay > 0) {
            runEndDelay = true;
        }

        return timer;
    };

    timer.start = function() {
        if (clock === null) {
            clock = setInterval(updateClock, 1000);
            clockRunning = true;
        }
        return timer;
    };

    timer.stopMain = function() {
        stopMain = true;
        return timer;
    };

    timer.stop = function() {
        if (clock !== null) {
            clearInterval(clock);
        }
        return timer;
    };

    timer.isDone = function() {
        clockRunning = false;
        return isDone;
    };

    function updateClock() {
        if (isCountdown) {
            seconds--;
            // Determine if the countdown has ended.
            if (seconds < 0) {
                if (runStartDelay === true) {
                    runStartDelay = false;
                    timerText.style("color", "black");
                    if (isIndeterminate === true) {
                        isCountdown = false;
                        seconds = 0;
                    } else {
                        seconds = recordDuration;
                    }
                    runMainTimer = true;
                } else if (runMainTimer === true) {
                    runMainTimer = false;
                    if (runEndDelay === false) {
                        seconds = 0;
                        timer.stop();
                        isDone = true;
                    } else {
                        seconds = endDelay;
                        timerText.style("color", "red");
                    }
                } else if (runEndDelay === true) {
                    seconds = 0;
                    timer.stop();
                    isDone = true;
                }
            }
        } else {
            seconds++;
            if (stopMain === true) {
                if (runEndDelay === true) {
                    isCountdown = true;
                    runMainTimer = false;
                    seconds = endDelay;
                    timerText.style("color", "red");
                }
            }
        }
        timerText.text(seconds);
    }

    return timer;
})();
